import sys
import math
import random
from faker import Faker

fake_zh = Faker('zh_CN')
fake_en = Faker('en_US')

rand_bool = lambda: random.random() < 0.5

rand_int8 = lambda: random.randint(-2 ** 7, 2 ** 7 - 1)
rand_uint8 = lambda: random.randint(0, 2 ** 8 - 1)
rand_int16 = lambda: random.randint(-2 ** 15, 2 ** 15 - 1)
rand_uint16 = lambda: random.randint(0, 2 ** 16 - 1)
rand_int32 = lambda: random.randint(-2 ** 31, 2 ** 31 - 1)
rand_uint32 = lambda: random.randint(0, 2 ** 32 - 1)
rand_int64 = lambda: random.randint(-2 ** 63, 2 ** 63 - 1)
rand_uint64 = lambda: random.randint(0, 2 ** 64 - 1)

rand_float_01 = lambda: random.random()
rand_float_11 = lambda: random.uniform(-1.0, 1.0)
rand_float_pos = lambda: math.exp(random.uniform(math.log(sys.float_info.min), math.log(sys.float_info.max)))
rand_float_zeropos = lambda: random.choice([0.0, rand_float_pos()])
rand_float_any = lambda: random.choice([-rand_float_pos(), 0.0, rand_float_pos()])
rand_float_inf = lambda: random.choice([rand_float_any(), float('inf'), float('-inf')])
rand_float_nan = lambda: random.choice([rand_float_any(), float('nan')])

rand_name_zh = lambda: fake_zh.name()  # Zhang Wei, Li Na
rand_name_en = lambda: fake_en.name()  # John Smith
rand_email = lambda: fake_en.email()  # john@example.com

rand_url = lambda: fake_en.url()  # https://example.com/
rand_domain = lambda: fake_en.domain_name()  # example.com
rand_ip = lambda: fake_en.ipv4()  # 192.168.1.1
rand_mac = lambda: fake_en.mac_address()  # 00:1B:44:11:3A:B7
rand_username = lambda: fake_en.user_name()  # john_doe
rand_password = lambda: fake_en.password()  # xK9#mP2$
rand_fname = lambda: fake_en.file_name()  # document.pdf
rand_fpath = lambda: fake_en.file_path()  # /home/user/doc.pdf

rand_date = lambda: fake_en.date_object()  # 2023-05-15
rand_time = lambda: fake_en.time_object().replace(microsecond=0)  # 14:30:25
rand_datetime = lambda: fake_en.date_time().replace(microsecond=0)  # datetime object

if __name__ == '__main__':
    pass

    current_module = sys.modules[__name__]
    rand_funcs = {
        name: obj for name, obj in vars(current_module).items()
        if name.startswith('rand_') and callable(obj)
    }

    for name, func in rand_funcs.items():
        print(f"\n{name}:")
        for _ in range(3):
            val = func()
            print(f"{type(val)} {[val]}")



import asyncio
import traceback

from gatling.utility.xprint import print_flush


class CoroutineExecutor():

    def __init__(self, max_workers=5, logfctn=print_flush):
        self.max_workers = max_workers
        self.logfctn = logfctn
        self.coroutine_tasks = []

    def submit(self, loop_func, *args, **kwargs):
        try:
            loop = asyncio.get_running_loop()
            use_existing_loop = True
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            use_existing_loop = False


        async def main():
            tasks = [asyncio.create_task(loop_func(*args, **kwargs)) for i in range(self.max_workers)]
            self.coroutine_tasks.extend(tasks)
            await asyncio.gather(*tasks, return_exceptions=True)

        if use_existing_loop:
            return loop.create_task(main())
        try:
            loop.run_until_complete(main())
        except Exception:
            self.logfctn(f"{loop_func.__name__} event loop exception:")
            self.logfctn(traceback.format_exc())
        finally:
            try:
                pending = asyncio.all_tasks(loop)
                for t in pending:
                    t.cancel()
                if pending:
                    loop.run_until_complete(asyncio.gather(*pending, return_exceptions=True))
            except Exception:
                self.logfctn(traceback.format_exc())
            finally:
                loop.close()
        self.coroutine_tasks.clear()


if __name__ == '__main__':
    pass



import unittest

from gatling.storage.g_dctn.memory_dctn import MemoryDctn

Key_A = "A"
Key_B = "B"
Key_C = "C"
Value_1_A = 1
Value_3_B = 3
Value_2_C = 2


class TestMemoryDctnBase(unittest.TestCase):
    """Unit tests for MemoryDctn get, set, pop methods."""

    # ==================== __getitem__ ====================

    def test_getitem_existing_key(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        self.assertEqual(d[Key_A], Value_1_A)

    def test_getitem_missing_key_raises_keyerror(self):
        d = MemoryDctn()
        with self.assertRaises(KeyError):
            _ = d[Key_A]

    # ==================== __setitem__ ====================

    def test_setitem_new_key(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertEqual(d[Key_A], Value_1_A)
        self.assertEqual(len(d), 1)

    def test_setitem_overwrite_key(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        d[Key_A] = Value_3_B
        self.assertEqual(d[Key_A], Value_3_B)
        self.assertEqual(len(d), 1)

    # ==================== __delitem__ ====================

    def test_delitem_existing_key(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        del d[Key_A]
        self.assertEqual(len(d), 0)
        self.assertNotIn(Key_A, d)

    def test_delitem_missing_key_raises_keyerror(self):
        d = MemoryDctn()
        with self.assertRaises(KeyError):
            del d[Key_A]

    # ==================== set ====================

    def test_set_new_key_returns_true(self):
        d = MemoryDctn()
        result = d.set(Key_A, Value_1_A)
        self.assertTrue(result)
        self.assertEqual(d[Key_A], Value_1_A)

    def test_set_existing_key_returns_false(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.set(Key_A, Value_3_B)
        self.assertFalse(result)
        self.assertEqual(d[Key_A], Value_3_B)

    # ==================== get ====================

    def test_get_existing_key(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        self.assertEqual(d.get(Key_A), Value_1_A)

    def test_get_missing_key_raises_keyerror(self):
        d = MemoryDctn()
        with self.assertRaises(KeyError):
            d.get(Key_A)

    def test_get_missing_key_with_default(self):
        d = MemoryDctn()
        result = d.get(Key_A, Value_2_C)
        self.assertEqual(result, Value_2_C)

    def test_get_missing_key_with_none_default(self):
        d = MemoryDctn()
        result = d.get(Key_A, None)
        self.assertIsNone(result)

    def test_get_missing_key_raises_custom_exception(self):
        d = MemoryDctn()
        with self.assertRaises(ValueError):
            d.get(Key_A, ValueError)

    # ==================== pop ====================

    def test_pop_existing_key(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.pop(Key_A)
        self.assertEqual(result, Value_1_A)
        self.assertNotIn(Key_A, d)

    def test_pop_missing_key_raises_keyerror(self):
        d = MemoryDctn()
        with self.assertRaises(KeyError):
            d.pop(Key_A)

    def test_pop_missing_key_with_default(self):
        d = MemoryDctn()
        result = d.pop(Key_A, Value_2_C)
        self.assertEqual(result, Value_2_C)

    def test_pop_missing_key_with_none_default(self):
        d = MemoryDctn()
        result = d.pop(Key_A, None)
        self.assertIsNone(result)

    def test_pop_missing_key_raises_custom_exception(self):
        d = MemoryDctn()
        with self.assertRaises(ValueError):
            d.pop(Key_A, ValueError)

    # ==================== setmany ====================

    def test_setmany_with_dict(self):
        d = MemoryDctn()
        count = d.setmany({Key_A: Value_1_A, Key_B: Value_3_B})
        self.assertEqual(count, 2)
        self.assertEqual(d[Key_A], Value_1_A)
        self.assertEqual(d[Key_B], Value_3_B)

    def test_setmany_with_kwargs(self):
        d = MemoryDctn()
        count = d.setmany(A=Value_1_A, B=Value_3_B)
        self.assertEqual(count, 2)
        self.assertEqual(d[Key_A], Value_1_A)
        self.assertEqual(d[Key_B], Value_3_B)

    def test_setmany_with_iterable(self):
        d = MemoryDctn()
        count = d.setmany([(Key_A, Value_1_A), (Key_B, Value_3_B)])
        self.assertEqual(count, 2)
        self.assertEqual(d[Key_A], Value_1_A)
        self.assertEqual(d[Key_B], Value_3_B)

    def test_setmany_with_dict_and_kwargs(self):
        d = MemoryDctn()
        count = d.setmany({Key_A: Value_1_A}, B=Value_3_B)
        self.assertEqual(count, 2)
        self.assertEqual(d[Key_A], Value_1_A)
        self.assertEqual(d[Key_B], Value_3_B)

    # ==================== getmany ====================

    def test_getmany_all_exist(self):
        d = MemoryDctn()
        d.setmany({Key_A: Value_1_A, Key_B: Value_3_B})
        result = d.getmany({Key_A: 0, Key_B: 0})
        self.assertEqual(result, {Key_A: Value_1_A, Key_B: Value_3_B})

    def test_getmany_with_defaults(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.getmany({Key_A: 0, Key_C: Value_2_C})
        self.assertEqual(result, {Key_A: Value_1_A, Key_C: Value_2_C})
        self.assertEqual(len(d), 1)  # unchanged

    def test_getmany_with_kwargs(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.getmany(A=0, C=Value_2_C)
        self.assertEqual(result, {Key_A: Value_1_A, Key_C: Value_2_C})

    def test_getmany_raises_on_missing(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        with self.assertRaises(KeyError):
            d.getmany({Key_A: 0, Key_C: KeyError})

    # ==================== popmany ====================

    def test_popmany_all_exist(self):
        d = MemoryDctn()
        d.setmany({Key_A: Value_1_A, Key_B: Value_3_B})
        result = d.popmany({Key_A: 0, Key_B: 0})
        self.assertEqual(result, {Key_A: Value_1_A, Key_B: Value_3_B})
        self.assertEqual(len(d), 0)

    def test_popmany_with_defaults(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.popmany({Key_A: 0, Key_C: Value_2_C})
        self.assertEqual(result, {Key_A: Value_1_A, Key_C: Value_2_C})
        self.assertEqual(len(d), 0)

    def test_popmany_with_kwargs(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        result = d.popmany(A=0, C=Value_2_C)
        self.assertEqual(result, {Key_A: Value_1_A, Key_C: Value_2_C})
        self.assertEqual(len(d), 0)

    def test_popmany_raises_on_missing(self):
        d = MemoryDctn()
        d.set(Key_A, Value_1_A)
        with self.assertRaises(KeyError):
            d.popmany({Key_A: 0, Key_C: KeyError})



from functools import update_wrapper


def combo_wraps(func, rename_rule=None):
    def decorator(wrapper):
        # copy all except __name__
        attrs = ('__module__', '__qualname__', '__annotations__', '__doc__')
        update_wrapper(wrapper, func, assigned=attrs)

        outter_wrapper_name = wrapper.__name__

        # custom rule
        if rename_rule is None:
            wrapper.__name__ = f"{outter_wrapper_name}({func.__name__})"
        else:
            wrapper.__name__ = rename_rule(outter_wrapper_name, func.__name__)

        return wrapper

    return decorator


def call_with(*args, **kwargs):
    def fmt_args_kwargs():
        parts = []
        if args:
            parts.append(", ".join(map(repr, args)))
        if kwargs:
            kw = ", ".join(f"{k}={v!r}" for k, v in kwargs.items())
            parts.append(kw)
        return ", ".join(parts)

    argstr = fmt_args_kwargs()

    def decorator(func):
        @combo_wraps(
            func,
            rename_rule=lambda wrapname, funcname: f"{wrapname}({argstr}).{funcname}"
        )
        def callwith():
            return func(*args, **kwargs)

        return callwith

    return decorator



import asyncio
import queue
import threading
import traceback
from concurrent.futures import Future
from typing import Callable, Optional, Any

from gatling.runtime.task_manager.coroutine_executor import CoroutineExecutor
from gatling.runtime.task_manager.runtime_task_manager_base import RuntimeTaskManager
from gatling.storage.g_queue.base_queue import BaseQueue
from gatling.storage.g_queue.memory_queue import MemoryQueue
from gatling.utility.xprint import print_flush


async def async_producer_iter_loop(fctn, qwait, qwork, qerrr, qdone, asyncio_stop_event, retry_on_error, retry_empty_interval, errlogfctn):
    while True:
        try:
            arg = qwait.get(block=False)
            qwork.put(arg)
            try:
                fut_iter = fctn(arg)
                async for item in fut_iter:
                    qdone.put(item)
            except Exception:
                errlogfctn(traceback.format_exc())
                qerrr.put(arg)
                if retry_on_error:
                    qwait.put(arg)
            finally:
                qwork.get(block=False)
        except queue.Empty:
            if asyncio_stop_event.is_set():
                break
            else:
                await asyncio.sleep(retry_empty_interval)


class RuntimeTaskManagerCoroutineIterator(RuntimeTaskManager):

    def __init__(self, fctn: Callable,
                 qwait: BaseQueue[Any],
                 qwork: BaseQueue[Future],
                 qerrr: BaseQueue[Any],
                 qdone: BaseQueue[Any],
                 worker: int = 1,
                 retry_on_error: bool = False,
                 retry_empty_interval=0.001,
                 errlogfctn=print_flush):
        super().__init__(fctn, qwait, qwork, qerrr, qdone, worker=worker,retry_on_error=retry_on_error,retry_empty_interval=retry_empty_interval)

        self.asyncio_stop_event: asyncio.Event = asyncio.Event()  # False
        self.asyncio_running_executor: Optional[CoroutineExecutor] = None
        self.errlogfctn = errlogfctn

        self.producers = []

    def __len__(self):
        return 0 if (self.asyncio_running_executor is None) else self.asyncio_running_executor.max_workers

    def __str__(self):
        return "CoIt" + super().__str__()

    def start(self, worker=None):

        if self.asyncio_running_executor is not None:
            raise RuntimeError(f"{str(self)} already started")
        if self.asyncio_stop_event.is_set():
            raise RuntimeError(f"{str(self)} is stopping")

        self.errlogfctn(f"{self} start triggered ... ")
        self.asyncio_running_executor = CoroutineExecutor(max_workers=worker, logfctn=self.errlogfctn)

        # start coroutine iterator
        producer_thread = threading.Thread(target=self.asyncio_running_executor.submit, args=(async_producer_iter_loop, self.fctn, self.qwait, self.qwork, self.qerrr, self.qdone, self.asyncio_stop_event,self.retry_on_error, self.retry_empty_interval, self.errlogfctn), daemon=True)
        producer_thread.start()
        self.producers.append(producer_thread)

        self.errlogfctn(f"{str(self)} started >>>")

    def stop(self):
        if self.asyncio_running_executor is None:
            return False
        if self.asyncio_stop_event.is_set():
            return False

        self.errlogfctn(f"{self} stop triggered ... ")
        self.asyncio_stop_event.set()

        for producer_thread in self.producers:
            producer_thread.join()
        self.producers.clear()

        self.asyncio_running_executor_worker = 0

        self.asyncio_stop_event.clear()

        self.errlogfctn(f"{str(self)} stopped !!!")
        return True


if __name__ == '__main__':
    pass
    from gatling.vtasks.sample_tasks import async_fake_iter_net

    rt = RuntimeTaskManagerCoroutineIterator(async_fake_iter_net, qwait=MemoryQueue(), qwork=MemoryQueue(), qerrr=MemoryQueue(), qdone=MemoryQueue())

    with rt.execute(worker=5, log_interval=1, logfctn=print_flush):
        for i in range(10):
            rt.qwait.put(i)

    print(f"[{len(rt.qdone)}] : {list(rt.qdone)}")



from collections import defaultdict

from pygments.lexers import data

from gatling.storage.g_table.table_ao_file_tsv import KEY_IDX
from gatling.utility.rand_tools import (
    rand_bool, rand_uint8, rand_int32, rand_float_01, rand_float_pos,
    rand_float_any, rand_float_inf, rand_name_zh,
    rand_name_en, rand_url, rand_ip, rand_username, rand_password,
    rand_fpath, rand_date, rand_time, rand_datetime
)

const_key2rand = {
    # str - account
    'account': rand_username,
    'secret': rand_password,

    # bool
    'is_active': rand_bool,

    # int
    'level': rand_uint8,
    'balance': rand_int32,

    # float
    'progress': rand_float_01,
    'price': rand_float_pos,
    'temperature': rand_float_any,
    'threshold_inf': rand_float_inf,

    # str - name
    'nickname_zh': rand_name_zh,
    'nickname_en': rand_name_en,

    # str - network
    'homepage': rand_url,
    'ip_address': rand_ip,

    # str - path
    'save_path': rand_fpath,

    # datetime
    'birthday': rand_date,
    'alarm': rand_time,
    'created_at': rand_datetime,
}

const_key2type = {key: type(rf()) for key, rf in const_key2rand.items()}
const_key2type_extra = {KEY_IDX: int, **const_key2type}
const_keys = list(const_key2rand.keys())
const_keys_extra =[KEY_IDX, *const_keys]

def rand_row():
    return {key: rf() for key, rf in const_key2rand.items()}


def filterbykeys(data, keys):
    if isinstance(data, dict):
        return {k: data[k] for k in keys}
    elif isinstance(data, list):
        return [filterbykeys(row, keys) for row in data]


def rows2cols(rows,keys):
    if isinstance(rows, dict):
        return rows
    elif isinstance(rows, list):
        if len(rows) == 0:
            return {k:[] for k in keys}
        else:
            res = defaultdict(list)
            for row in rows:
                for k, v in row.items():
                    res[k].append(v)
            return dict(res)


if __name__ == '__main__':
    pass
    data = [
        {'a': 1, 'b': 2},
        {'a': 3, 'b': 4},
        {'a': 5, 'b': 6}
    ]

    print(rows2cols(data))



import unittest

from gatling.storage.g_dctn.memory_dctn import MemoryDctn

Key_A = "A"
Key_B = "B"
Key_C = "C"
Value_1_A = 1
Value_3_B = 3
Value_2_C = 2


class TestMemoryDctnMore(unittest.TestCase):
    """Unit tests for MemoryDctn class."""

    def test_basic_set_get_len(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertEqual(len(d), 1)
        self.assertEqual(d[Key_A], Value_1_A)

    def test_basic_set_pop_len(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertEqual(len(d), 1)
        self.assertEqual(d.pop(Key_A), Value_1_A)
        self.assertEqual(len(d), 0)

    def test_basic_get_default(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertEqual(d.get(Key_C, Value_2_C), Value_2_C)

    def test_basic_pop_default(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertEqual(d.pop(Key_C, Value_2_C), Value_2_C)
        self.assertEqual(len(d), 1)

    def test_clear(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        d[Key_B] = Value_3_B
        self.assertEqual(len(d), 2)
        d.clear()
        self.assertEqual(len(d), 0)

    def test_del(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        del d[Key_A]
        self.assertEqual(len(d), 0)
        with self.assertRaises(KeyError):
            print(d[Key_A])

    def test_keys(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        d[Key_B] = Value_3_B
        self.assertEqual(list(d.keys()), [Key_A, Key_B])

    def test_values(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        d[Key_B] = Value_3_B
        self.assertEqual(list(d.values()), [Value_1_A, Value_3_B])

    def test_items(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        d[Key_B] = Value_3_B
        self.assertEqual(list(d.items()), [(Key_A, Value_1_A), (Key_B, Value_3_B)])

    def test_contains(self):
        d = MemoryDctn()
        d[Key_A] = Value_1_A
        self.assertIn(Key_A, d)
        self.assertNotIn(Key_B, d)

    def test_eq(self):
        d1 = MemoryDctn()
        d1.setmany({Key_A: Value_1_A, Key_B: Value_3_B})
        d2 = MemoryDctn()
        d2.setmany({Key_B: Value_3_B, Key_A: Value_1_A})
        self.assertEqual(d1, d2)
        d3 = {Key_B: Value_3_B, Key_A: Value_1_A}
        self.assertEqual(d1, d3)

    def test_ne(self):
        d1 = MemoryDctn()
        d1.setmany({Key_A: Value_1_A, Key_B: Value_3_B})
        d2 = MemoryDctn()
        d2.setmany({Key_B: Value_3_B, Key_A: Value_2_C})
        self.assertNotEqual(d1, d2)
        d3 = {Key_B: Value_3_B, Key_A: Value_2_C}
        self.assertNotEqual(d1, d3)

    def test_sort(self):
        d = MemoryDctn()
        d[Key_C] = Value_2_C
        d[Key_A] = Value_1_A
        d[Key_B] = Value_3_B
        self.assertEqual(list(d.keys()), [Key_C, Key_A, Key_B])
        d.sort()
        self.assertEqual(list(d.keys()), [Key_A, Key_C, Key_B])
        d.sort(reverse=True)
        self.assertEqual(list(d.keys()), [Key_B, Key_C, Key_A])
        d.sort(by='key')
        self.assertEqual(list(d.keys()), [Key_A, Key_B, Key_C])
        d.sort(by='key', reverse=True)
        self.assertEqual(list(d.keys()), [Key_C, Key_B, Key_A])
        d.sort(by=lambda k, v: {Key_B: 0, Key_A: 1, Key_C: 2}[k])
        self.assertEqual(list(d.keys()), [Key_B, Key_A, Key_C])


    def test_open_and_close(self):
        d = MemoryDctn()
        try:
            d.open()
            d.close()
        except Exception as e:
            self.fail(f"open/close raised unexpectedly: {e}")

    def test_context_manager(self):
        d = MemoryDctn()
        try:
            with d:
                pass
        except Exception as e:
            self.fail(f"context manager raised unexpectedly: {e}")



from typing import Callable

from gatling.utility.decorator_tools import call_with
from gatling.utility.watch import Watch
from gatling.vtasks.a_const import cpu_flops_per_sec, flops_target


def calc_fctn_flops(fctn: Callable, init_flops, flops_per_sec=cpu_flops_per_sec, num=1, xrange=range):
    w = Watch()
    for i in xrange(num):
        fctn()
    total_cost_secs = w.see_seconds()
    percall_cost_secs = total_cost_secs / num
    percall_flops = flops_per_sec * percall_cost_secs

    return percall_flops


def approx_calc_fctn_flops(fctn: Callable, init_flops, iter, flops_per_sec=cpu_flops_per_sec, xrange=range, flops=flops_target, name=''):
    print(f"{fctn.__name__ if not name else name}.flops : {init_flops:.0f}", end='')
    percall_flops = init_flops
    for i in xrange(iter):
        num = int(flops / percall_flops) + 1
        percall_flops: int = calc_fctn_flops(fctn=fctn, init_flops=percall_flops, flops_per_sec=flops_per_sec, num=num, xrange=xrange)
        print(f" => {percall_flops:.0f}", end='')
    print()
    return percall_flops


################################################################################################################################################

if __name__ == '__main__':
    pass
    from gatling.vtasks.c_base_fctns import random_seed_int, mix64

    approx_calc_fctn_flops(random_seed_int, init_flops=100, iter=10)
    approx_calc_fctn_flops(call_with(10)(mix64), init_flops=100, iter=10)



import random


def random_seed_int():
    return random.getrandbits(64)


def mix64(x: int) -> int:
    x = (x ^ (x >> 30)) * 0xbf58476d1ce4e5b9
    x = (x ^ (x >> 27)) * 0x94d049bb133111eb
    x = x ^ (x >> 31)
    return x & 0xFFFFFFFFFFFFFFFF



import json
import unittest
from gatling.utility.mem_tools import sizeof
from helper.get_varname import get_var_name
from storage.g_dctn import test_memory_dctn_str_0_objs
from gatling.storage.g_dctn.memory_dctn import MemoryDctn
from helper.dynamic_testcase import DynamicTestCase

varname2obj = {get_var_name(obj, vars(test_memory_dctn_str_0_objs)): obj for obj in test_memory_dctn_str_0_objs.memorydctn_objs}


# DynamicTestCase.set_name('TestRuntimeTaskManagerCoroutine')
class TestMemoryDctnStr(DynamicTestCase):
    pass


# Define Test Case Function
def testcase_fctn(cand: dict, level):
    md = MemoryDctn()

    md.setmany(cand)
    sent = md.str(level)
    # sent = str(md)
    print(f"===candidate===[{sizeof(cand)}]")
    print(cand)
    print(f"===serialized[{sizeof(md)}]===")
    print(sent)
    print("===end===\n")
    back = json.loads(sent)

    assert md == back, f"{sent} != {back}"


# === Dynamic Register Test Case ===
for varname, obj in varname2obj.items():
    for level in [-1, 0, 1, 2, 3]:
        testcase_name = f"test_{varname}_{level=}"
        TestMemoryDctnStr.append_testcase(testcase_name, testcase_fctn, obj, level)

if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest
import os
import tempfile
import orjson
import sys
from io import StringIO
from gatling.utility.io_fctns import read_json, save_json, read_jsonl, save_jsonl


class TestIOFunctions(unittest.TestCase):
    """Test read/save functions for JSON and JSONL files"""

    def setUp(self):
        self.tempdir = tempfile.TemporaryDirectory()
        self.dpath = self.tempdir.name
        self._stderr = sys.stderr
        sys.stderr = StringIO()
        self._stdout = sys.stdout
        sys.stdout = StringIO()

    def tearDown(self):
        self.tempdir.cleanup()
        sys.stderr = self._stderr
        sys.stdout = self._stdout

    # ==================== read_json ====================
    def test_read_json_simple(self):
        """Test reading a simple JSON file"""
        fpath = os.path.join(self.dpath, 'simple.json')
        data = {'name': 'test', 'value': 123}
        with open(fpath, 'wb') as f:
            f.write(orjson.dumps(data))

        result = read_json(fpath)
        self.assertEqual(result, data)

    def test_read_json_nested(self):
        """Test reading nested JSON structure"""
        fpath = os.path.join(self.dpath, 'nested.json')
        data = {
            'user': {
                'name': 'Alice',
                'address': {'city': 'New York', 'zip': '10001'}
            },
            'tags': ['a', 'b', 'c']
        }
        with open(fpath, 'wb') as f:
            f.write(orjson.dumps(data))

        result = read_json(fpath)
        self.assertEqual(result, data)

    def test_read_json_array(self):
        """Test reading JSON array as root"""
        fpath = os.path.join(self.dpath, 'array.json')
        data = [1, 2, 3, {'a': 1}]
        with open(fpath, 'wb') as f:
            f.write(orjson.dumps(data))

        result = read_json(fpath)
        self.assertEqual(result, data)

    def test_read_json_special_types(self):
        """Test reading JSON with null, bool, numbers"""
        fpath = os.path.join(self.dpath, 'special.json')
        data = {'null_val': None, 'bool_true': True, 'bool_false': False, 'float': 3.14, 'negative': -100}
        with open(fpath, 'wb') as f:
            f.write(orjson.dumps(data))

        result = read_json(fpath)
        self.assertEqual(result, data)

    def test_read_json_unicode(self):
        """Test reading JSON with unicode characters"""
        fpath = os.path.join(self.dpath, 'unicode.json')
        data = {'name': 'Alice', 'emoji': '😀🎉', 'chinese': '你好'}
        with open(fpath, 'wb') as f:
            f.write(orjson.dumps(data))

        result = read_json(fpath)
        self.assertEqual(result, data)

    def test_read_json_empty_object(self):
        """Test reading empty JSON object"""
        fpath = os.path.join(self.dpath, 'empty_obj.json')
        with open(fpath, 'wb') as f:
            f.write(b'{}')

        result = read_json(fpath)
        self.assertEqual(result, {})

    def test_read_json_empty_array(self):
        """Test reading empty JSON array"""
        fpath = os.path.join(self.dpath, 'empty_arr.json')
        with open(fpath, 'wb') as f:
            f.write(b'[]')

        result = read_json(fpath)
        self.assertEqual(result, [])

    def test_read_json_empty_file(self):
        """Test reading empty file returns None"""
        fpath = os.path.join(self.dpath, 'empty.json')
        with open(fpath, 'wb') as _:
            pass

        result = read_json(fpath)
        self.assertIsNone(result)

    def test_read_json_invalid(self):
        """Test reading invalid JSON returns None"""
        fpath = os.path.join(self.dpath, 'invalid.json')
        with open(fpath, 'wb') as f:
            f.write(b'{invalid json}')

        result = read_json(fpath)
        self.assertIsNone(result)

    def test_read_json_not_found(self):
        """Test reading non-existent file raises FileNotFoundError"""
        fpath = os.path.join(self.dpath, 'not_exist.json')
        with self.assertRaises(FileNotFoundError):
            read_json(fpath)

    # ==================== save_json ====================
    def test_save_json_simple(self):
        """Test saving a simple JSON file"""
        fpath = os.path.join(self.dpath, 'save_simple.json')
        data = {'name': 'test', 'value': 456}

        result = save_json(data, fpath)
        self.assertTrue(result)
        self.assertTrue(os.path.exists(fpath))
        self.assertEqual(read_json(fpath), data)

    def test_save_json_nested(self):
        """Test saving nested JSON structure"""
        fpath = os.path.join(self.dpath, 'save_nested.json')
        data = {'level1': {'level2': {'level3': [1, 2, 3]}}}

        result = save_json(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), data)

    def test_save_json_array(self):
        """Test saving JSON array as root"""
        fpath = os.path.join(self.dpath, 'save_array.json')
        data = [{'id': 1}, {'id': 2}]

        result = save_json(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), data)

    def test_save_json_special_types(self):
        """Test saving JSON with null, bool, numbers"""
        fpath = os.path.join(self.dpath, 'save_special.json')
        data = {'null_val': None, 'bool_true': True, 'bool_false': False, 'float': 3.14159, 'negative': -100}

        result = save_json(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), data)

    def test_save_json_unicode(self):
        """Test saving JSON with unicode characters"""
        fpath = os.path.join(self.dpath, 'save_unicode.json')
        data = {'name': 'Bob', 'emoji': '🎉', 'japanese': 'こんにちは'}

        result = save_json(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), data)

    def test_save_json_empty_object(self):
        """Test saving empty JSON object"""
        fpath = os.path.join(self.dpath, 'save_empty_obj.json')

        result = save_json({}, fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), {})

    def test_save_json_empty_array(self):
        """Test saving empty JSON array"""
        fpath = os.path.join(self.dpath, 'save_empty_arr.json')

        result = save_json([], fpath)
        self.assertTrue(result)
        self.assertEqual(read_json(fpath), [])

    def test_save_json_indent(self):
        """Test saving JSON with indentation"""
        fpath = os.path.join(self.dpath, 'save_indent.json')
        data = {'a': 1, 'b': 2}

        result = save_json(data, fpath, indent=True)
        self.assertTrue(result)

        with open(fpath, 'rb') as f:
            content = f.read()
        self.assertIn(b'\n', content)

    def test_save_json_overwrite(self):
        """Test overwriting existing file"""
        fpath = os.path.join(self.dpath, 'overwrite.json')
        save_json({'old': 1}, fpath)
        save_json({'new': 2}, fpath)

        result = read_json(fpath)
        self.assertEqual(result, {'new': 2})

    def test_save_json_invalid_path(self):
        """Test saving to invalid path returns False"""
        fpath = '/invalid_path/no_permission/test.json'
        result = save_json({'a': 1}, fpath)
        self.assertFalse(result)

    # ==================== read_jsonl ====================
    def test_read_jsonl_simple(self):
        """Test reading a simple JSONL file"""
        fpath = os.path.join(self.dpath, 'simple.jsonl')
        data = [{'id': 1}, {'id': 2}, {'id': 3}]
        with open(fpath, 'wb') as f:
            f.write(b'\n'.join(orjson.dumps(d) for d in data))

        result = read_jsonl(fpath)
        self.assertEqual(result, data)

    def test_read_jsonl_single_line(self):
        """Test reading JSONL with single line"""
        fpath = os.path.join(self.dpath, 'single.jsonl')
        with open(fpath, 'wb') as f:
            f.write(b'{"id": 1}')

        result = read_jsonl(fpath)
        self.assertEqual(result, [{'id': 1}])

    def test_read_jsonl_nested(self):
        """Test reading JSONL with nested objects"""
        fpath = os.path.join(self.dpath, 'nested.jsonl')
        data = [{'user': {'name': 'Alice', 'info': {'age': 20}}}, {'user': {'name': 'Bob', 'info': {'age': 30}}}]
        with open(fpath, 'wb') as f:
            f.write(b'\n'.join(orjson.dumps(d) for d in data))

        result = read_jsonl(fpath)
        self.assertEqual(result, data)

    def test_read_jsonl_special_types(self):
        """Test reading JSONL with null, bool, numbers"""
        fpath = os.path.join(self.dpath, 'special.jsonl')
        data = [{'val': None}, {'val': True}, {'val': False}, {'val': 3.14}, {'val': -100}]
        with open(fpath, 'wb') as f:
            f.write(b'\n'.join(orjson.dumps(d) for d in data))

        result = read_jsonl(fpath)
        self.assertEqual(result, data)

    def test_read_jsonl_unicode(self):
        """Test reading JSONL with unicode characters"""
        fpath = os.path.join(self.dpath, 'unicode.jsonl')
        data = [{'name': 'Alice', 'emoji': '😀'}, {'name': 'Bob', 'chinese': '你好'}]
        with open(fpath, 'wb') as f:
            f.write(b'\n'.join(orjson.dumps(d) for d in data))

        result = read_jsonl(fpath)
        self.assertEqual(result, data)

    def test_read_jsonl_empty_lines(self):
        """Test reading JSONL with empty lines"""
        fpath = os.path.join(self.dpath, 'empty_lines.jsonl')
        with open(fpath, 'wb') as f:
            f.write(b'{"id": 1}\n\n{"id": 2}\n\n')

        result = read_jsonl(fpath)
        self.assertEqual(len(result), 2)

    def test_read_jsonl_empty_file(self):
        """Test reading empty JSONL file"""
        fpath = os.path.join(self.dpath, 'empty.jsonl')
        with open(fpath, 'wb') as _:
            pass

        result = read_jsonl(fpath)
        self.assertEqual(result, [])

    def test_read_jsonl_partial_invalid(self):
        """Test reading JSONL with some invalid lines"""
        fpath = os.path.join(self.dpath, 'partial_invalid.jsonl')
        with open(fpath, 'wb') as f:
            f.write(b'{"id": 1}\n{invalid}\n{"id": 3}')

        result = read_jsonl(fpath)
        self.assertEqual(len(result), 2)

    def test_read_jsonl_many_lines(self):
        """Test reading JSONL with many lines"""
        fpath = os.path.join(self.dpath, 'many.jsonl')
        data = [{'id': i} for i in range(1000)]
        with open(fpath, 'wb') as f:
            f.write(b'\n'.join(orjson.dumps(d) for d in data))

        result = read_jsonl(fpath)
        self.assertEqual(len(result), 1000)
        self.assertEqual(result[0], {'id': 0})
        self.assertEqual(result[-1], {'id': 999})

    def test_read_jsonl_not_found(self):
        """Test reading non-existent JSONL raises FileNotFoundError"""
        fpath = os.path.join(self.dpath, 'not_exist.jsonl')
        with self.assertRaises(FileNotFoundError):
            read_jsonl(fpath)

    # ==================== save_jsonl ====================
    def test_save_jsonl_simple(self):
        """Test saving a simple JSONL file"""
        fpath = os.path.join(self.dpath, 'save_simple.jsonl')
        data = [{'id': 1}, {'id': 2}, {'id': 3}]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertTrue(os.path.exists(fpath))
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_single_item(self):
        """Test saving JSONL with single item"""
        fpath = os.path.join(self.dpath, 'save_single.jsonl')
        data = [{'id': 1}]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_nested(self):
        """Test saving JSONL with nested objects"""
        fpath = os.path.join(self.dpath, 'save_nested.jsonl')
        data = [{'user': {'info': {'age': 20}}}, {'user': {'info': {'age': 30}}}]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_special_types(self):
        """Test saving JSONL with null, bool, numbers"""
        fpath = os.path.join(self.dpath, 'save_special.jsonl')
        data = [{'val': None}, {'val': True}, {'val': False}, {'val': -99.99}]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_unicode(self):
        """Test saving JSONL with unicode characters"""
        fpath = os.path.join(self.dpath, 'save_unicode.jsonl')
        data = [{'name': 'Charlie', 'korean': '안녕하세요'}, {'name': 'Diana', 'emoji': '🚀'}]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_empty(self):
        """Test saving empty JSONL"""
        fpath = os.path.join(self.dpath, 'save_empty.jsonl')

        result = save_jsonl([], fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), [])

    def test_save_jsonl_many_items(self):
        """Test saving JSONL with many items"""
        fpath = os.path.join(self.dpath, 'save_many.jsonl')
        data = [{'id': i, 'value': f'item_{i}'} for i in range(1000)]

        result = save_jsonl(data, fpath)
        self.assertTrue(result)
        self.assertEqual(read_jsonl(fpath), data)

    def test_save_jsonl_overwrite(self):
        """Test overwriting existing JSONL file"""
        fpath = os.path.join(self.dpath, 'overwrite.jsonl')
        save_jsonl([{'old': 1}], fpath)
        save_jsonl([{'new': 2}], fpath)

        result = read_jsonl(fpath)
        self.assertEqual(result, [{'new': 2}])

    def test_save_jsonl_invalid_path(self):
        """Test saving to invalid path returns False"""
        fpath = '/invalid_path/no_permission/test.jsonl'
        result = save_jsonl([{'a': 1}], fpath)
        self.assertFalse(result)


if __name__ == '__main__':
    unittest.main(verbosity=2)


import unittest
from datetime import timedelta
import time

from gatling.utility.watch import Watch, watch_time  # adjust import path if needed


class TestWatchTime(unittest.TestCase):
    """Functional tests for Watch class and @watch_time decorator (no timing checks)."""

    # ------------------------------------------------------------
    def test_watch_time_decorator(self):
        """Ensure @watch_time executes and returns expected value."""

        @watch_time
        def decorated_function(value):
            """Simple test function that returns its argument."""
            # a small sleep just to simulate work
            time.sleep(0.01)
            return f"OK-{value}"

        result = decorated_function("demo")

        # Function runs successfully
        self.assertIsInstance(result, str)
        self.assertTrue(result.startswith("OK-"))

    # ------------------------------------------------------------
    def test_watch_class_methods(self):
        """Ensure Watch methods run and return correct types."""

        watch = Watch()

        # see_timedelta() should return timedelta
        td = watch.see_timedelta()
        self.assertIsInstance(td, timedelta)

        # see_seconds() should return float
        secs = watch.see_seconds()
        self.assertIsInstance(secs, float)

        # total_timedelta() should return timedelta
        total_td = watch.total_timedelta()
        self.assertIsInstance(total_td, timedelta)

        # total_seconds() should return float
        total_secs = watch.total_seconds()
        self.assertIsInstance(total_secs, float)

        # Internal structure check
        self.assertTrue(hasattr(watch, "records"))
        self.assertIsInstance(watch.records, list)

        print("\nAll Watch methods executed and returned correct types.")


if __name__ == "__main__":
    unittest.main(verbosity=2)



import unittest


class DynamicTestCase(unittest.TestCase):

    @classmethod
    def set_name(cls, name):
        cls.__name__ = name

    @classmethod
    def append_testcase(cls, test_case_name, test_case_fctn, *args, **kwargs):
        """append a test case to the class"""
        if not test_case_name.startswith("test_"):
            test_case_name = f"test_{test_case_name}"

        def testcase_method(self):
            return test_case_fctn(*args, **kwargs)
        testcase_method.__name__ = test_case_name
        setattr(cls, test_case_name, testcase_method)


if __name__ == '__main__':
    pass
    DynamicTestCase.set_name('MyTestCase')
    # or
    class MyTestCase(DynamicTestCase):
        pass



import unittest
import os
import tempfile
from gatling.utility.proxy_mana import ProxyManager
from gatling.utility.io_fctns import read_json, save_json


class TestProxyManager(unittest.TestCase):
    """Test the initialization, saving, and random proxy functionality of ProxyManager"""

    def setUp(self):
        # Create a temporary directory
        self.tempdir = tempfile.TemporaryDirectory()
        self.dpath_base_asset = self.tempdir.name

        # Create a mock data.http.json file
        self.http_json_path = os.path.join(self.dpath_base_asset, 'data.http.json')
        proxyinfos = [
            {'proxy': 'http://proxy1.example.com:8080'},
            {'proxy': 'http://proxy2.example.com:8080'}
        ]
        save_json(proxyinfos, self.http_json_path)  # Use your own save_json()

        # Mock a preset object
        class Preset:
            pass

        self.preset = Preset()
        self.preset.dpath_base_asset = self.dpath_base_asset

    def tearDown(self):
        # Automatically clean up the temporary directory
        self.tempdir.cleanup()

    def test_proxy_manager_workflow(self):
        """Test the full workflow of ProxyManager"""
        # Read data.http.json (use your own read_json)
        proxyinfos = read_json(self.http_json_path)
        proxies = [pxinfo['proxy'] for pxinfo in proxyinfos]

        # Initialize ProxyManager
        fpath_proxyinfo = os.path.join(self.preset.dpath_base_asset, 'proxyinfos.http.jsonl')
        pm = ProxyManager(fpath_proxyinfo)
        pm.initialize(proxies)
        pm.save()

        # Verify that the file was saved successfully
        self.assertTrue(os.path.exists(fpath_proxyinfo), "proxyinfos.http.jsonl file not created")

        # Verify that random proxies are valid
        for _ in range(10):
            proxy = pm.rand_ts_proxy_for_aiohttp()
            self.assertIn(proxy, proxies, f"Returned proxy {proxy} not in initialization list")

        print("\n ProxyManager test passed")


if __name__ == '__main__':
    unittest.main(verbosity=2)



from abc import abstractmethod


class BaseTableAO():

    def __init__(self):
        super().__init__()

    @abstractmethod
    def append(self, row):
        pass

    @abstractmethod
    def extend(self, iterable):
        pass

    @abstractmethod
    def delete(self):
        pass

    @abstractmethod
    def clear(self):
        pass

    @abstractmethod
    def keys(self):
        pass

    @abstractmethod
    def rows(self):
        pass

    def col(self, key):
        pass

    @abstractmethod
    def __getitem__(self, slice: slice):
        pass

    @abstractmethod
    def __len__(self):
        pass

    @abstractmethod
    def __iter__(self):
        pass

    def __repr__(self):
        return f"<{self.__class__.__name__}({len(self)}x{len(self.keys())})-{id(self)}>"

    @abstractmethod
    def __enter__(self):
        pass

    @abstractmethod
    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @abstractmethod
    def exists(self) -> bool:
        pass

    @abstractmethod
    def pop(self) -> dict:
        pass

    @abstractmethod
    def shrink(self, n: int) -> list:
        pass


if __name__ == '__main__':
    pass



"""
Dump utilities for nested data structures.
- dumps_soft: collapse large objects to (count)[size] summary
- dumps_hard: indent up to level, flatten deeper content into single line
"""

import json
from typing import Any

from gatling.utility.mem_tools import sizeof


def _get_summary(obj, shown: int = None, trailing: bool = False) -> str:
    """
    Return summary for object.

    Args:
        obj: object to summarize
        shown: number of items already shown (None for collapsed value)
        trailing: if True, this is trailing summary for a list/dict (no type brackets needed)

    Returns:
        - dict: {...(shown/total)[size]...} or ...(shown/total)[size]... if trailing
        - list: [...(shown/total)[size]...] or ...(shown/total)[size]... if trailing
        - str: (len)[size]
        - other: [size]
    """
    size = sizeof(obj)
    if isinstance(obj, dict):
        count_str = f"...({shown if shown is not None else 0}/{len(obj)})[{size}]..."
        if trailing:
            return count_str
        return "{" + count_str + "}"
    elif isinstance(obj, (list, tuple)):
        count_str = f"...({shown if shown is not None else 0}/{len(obj)})[{size}]..."
        if trailing:
            return count_str
        bracket_open = "[" if isinstance(obj, list) else "("
        bracket_close = "]" if isinstance(obj, list) else ")"
        return bracket_open + count_str + bracket_close
    elif isinstance(obj, str):
        return f"({len(obj)})[{size}]"
    else:
        return f"[{size}]"


def _should_collapse(obj, max_items: int, max_size: int) -> bool:
    """Check if object should be collapsed to summary"""
    size = sizeof(obj)
    if size > max_size:
        return True
    if isinstance(obj, dict) and len(obj) > max_items:
        return True
    if isinstance(obj, (list, tuple)) and len(obj) > max_items:
        return True
    return False


def dumps_soft(
        obj: Any,
        level: int = -1,
        indent: int = 2,
        max_items: int = 10,
        max_size: int = 1000,
        show_items: int = 3,
        _depth: int = 0
) -> str:
    """
    Soft dump: collapse large objects to (count)[size] summary.

    Args:
        obj: object to dump
        level: max indent depth (-1 for unlimited, 0 for single line)
        indent: spaces per indent level
        max_items: collapse if item count exceeds this
        max_size: collapse if byte size exceeds this
        show_items: show first N items before collapsing (0 = show nothing, capped at max_items)
        _depth: internal use for recursion

    Returns:
        formatted string representation
    """
    # cap show_items at max_items
    show_items = min(show_items, max_items)

    ind = " " * (indent * _depth)
    ind_next = " " * (indent * (_depth + 1))

    # level 0: single line
    if level == 0:
        if _should_collapse(obj, max_items, max_size):
            return _get_summary(obj)
        return json.dumps(obj, ensure_ascii=False)

    # beyond depth limit
    if level > 0 and _depth >= level:
        if _should_collapse(obj, max_items, max_size):
            return _get_summary(obj)
        return json.dumps(obj, ensure_ascii=False)

    if isinstance(obj, dict):
        if not obj:
            return "{}"

        should_collapse_dict = _should_collapse(obj, max_items, max_size)
        items = []
        shown_count = 0

        for k, v in obj.items():
            # if dict should collapse and we've shown enough
            if should_collapse_dict and shown_count >= show_items:
                break

            key_str = json.dumps(k, ensure_ascii=False)
            if _should_collapse(v, max_items, max_size):
                val_str = _get_summary(v)  # collapsed value: include type brackets
            else:
                val_str = dumps_soft(v, level, indent, max_items, max_size, show_items, _depth + 1)
            items.append(f"{ind_next}{key_str}: {val_str}")
            shown_count += 1

        # add collapse summary if not all shown (trailing: no type brackets needed)
        if shown_count < len(obj):
            items.append(f"{ind_next}{_get_summary(obj, shown_count, trailing=True)}")

        return "{\n" + ",\n".join(items) + f"\n{ind}}}"

    elif isinstance(obj, (list, tuple)):
        if not obj:
            return "[]" if isinstance(obj, list) else "()"

        bracket_open = "[" if isinstance(obj, list) else "("
        bracket_close = "]" if isinstance(obj, list) else ")"

        should_collapse_list = _should_collapse(obj, max_items, max_size)
        items = []
        shown_count = 0

        for i, item in enumerate(obj):
            # if list should collapse and we've shown enough
            if should_collapse_list and shown_count >= show_items:
                break

            if _should_collapse(item, max_items, max_size):
                # collapsed item: show index and type brackets
                val_str = f"#{i}: {_get_summary(item)}"
            else:
                # full item: no index, keep original structure
                val_str = dumps_soft(item, level, indent, max_items, max_size, show_items, _depth + 1)
            items.append(f"{ind_next}{val_str}")
            shown_count += 1

        # add collapse summary if not all shown (trailing: no type brackets needed)
        if shown_count < len(obj):
            items.append(f"{ind_next}{_get_summary(obj, shown_count, trailing=True)}")

        return f"{bracket_open}\n" + ",\n".join(items) + f"\n{ind}{bracket_close}"

    else:
        return json.dumps(obj, ensure_ascii=False)


def dumps_hard(
        obj: Any,
        level: int = -1,
        indent: int = 2,
        max_items: int = None,
        max_size: int = None,
        show_items: int = None,
        _depth: int = 0
) -> str:
    """
    Hard dump: indent up to level, flatten deeper content into single line.

    Args:
        obj: object to dump
        level: max indent depth (-1 for unlimited, 0 for single line)
        indent: spaces per indent level
        max_items: ignored (for interface compatibility)
        max_size: ignored (for interface compatibility)
        show_items: ignored (for interface compatibility)
        _depth: internal use for recursion

    Returns:
        formatted JSON string
    """
    if level == 0:
        return json.dumps(obj, ensure_ascii=False)

    if level == -1:
        return json.dumps(obj, ensure_ascii=False, indent=indent)

    if _depth >= level:
        return json.dumps(obj, ensure_ascii=False)

    ind = " " * (indent * _depth)
    ind_next = " " * (indent * (_depth + 1))

    if isinstance(obj, dict):
        if not obj:
            return "{}"
        items = []
        for k, v in obj.items():
            key_str = json.dumps(k, ensure_ascii=False)
            val_str = dumps_hard(v, level, indent, _depth=_depth + 1)
            items.append(f"{ind_next}{key_str}: {val_str}")
        return "{\n" + ",\n".join(items) + f"\n{ind}}}"

    elif isinstance(obj, (list, tuple)):
        if not obj:
            return "[]" if isinstance(obj, list) else "()"
        bracket_open = "[" if isinstance(obj, list) else "("
        bracket_close = "]" if isinstance(obj, list) else ")"
        items = []
        for item in obj:
            val_str = dumps_hard(item, level, indent, _depth=_depth + 1)
            items.append(f"{ind_next}{val_str}")
        return f"{bracket_open}\n" + ",\n".join(items) + f"\n{ind}{bracket_close}"

    else:
        return json.dumps(obj, ensure_ascii=False)


if __name__ == "__main__":
    test_data = {
        "small": {"a": 1, "b": 2},
        "big_list": list(range(100)),
        "nested": {
            "level1": {
                "level2": {
                    "huge": list(range(1000))
                }
            }
        },
        "users": [
            {"name": "Alice", "scores": list(range(50))},
            {"name": "Bob", "scores": list(range(50))},
            {"name": "Charlie", "scores": list(range(50))},
            {"name": "David", "scores": list(range(50))},
            {"name": "Eve", "scores": list(range(50))},
        ],
        "simple_list": [1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12],
    }

    print("=== dumps_soft show_items=3 ===")
    print(dumps_soft(test_data, level=3, max_items=5, max_size=500, show_items=3))

    print("\n=== dumps_soft show_items=0 (no items, keep brackets) ===")
    print(dumps_soft(test_data, level=3, max_items=5, max_size=500, show_items=0))

    print("\n=== List example: users ===")
    print(dumps_soft(test_data["users"], level=2, max_items=3, max_size=500, show_items=2))

    print("\n=== List example: big_list ===")
    print(dumps_soft(test_data["big_list"], level=2, max_items=5, max_size=200, show_items=3))



# Basic types
obj_basic = {
    "string": "hello world",
    "string_cn": "中文测试",
    "int": 42,
    "int_neg": -100,
    "int_zero": 0,
    "int_big": 10 ** 18,
    "float": 3.14159,
    "float_neg": -2.718,
    "float_exp": 1.23e-10,
    "float_zero": 0.0,
    "bool_true": True,
    "bool_false": False,
    "null": None,
}

# String edge cases
obj_string_edge = {
    "string_escape": "line1\nline2\ttab\"quote'单引号",
    "string_tricky": "It's True! None of them, False alarm",
    "string_backslash": "C:\\Users\\test",
    "string_unicode": "emoji 🎉 and \u0000 null",
    "string_json_like": '{"not": "real"}',
    "empty_str": "",
}

# Empty containers
obj_empty = {
    "empty_dict": {},
    "empty_list": [],
    "empty_str": "",
}

# Nested structures
obj_nested = {
    "nested_dict": {
        "level2": {
            "level3": {
                "level4": {"deep": "value"}
            }
        }
    },
    "nested_list": [[[["deep"]]]],
    "mixed_nest": {
        "users": [
            {"name": "Alice", "scores": [95, 87, 92]},
            {"name": "Bob", "scores": [88, 91, 85]}
        ]
    },
}

# Special keys
obj_special_keys = {
    "": "empty key",
    "key with space": "space",
    "key\"quote": "has quote",
    "key'single": "single quote",
    "key\nnewline": "newline in key",
    "123": "numeric string key",
    "True": "key looks like bool",
    "None": "key looks like null",
    "null": "key is literal null word",
}

# Tricky string values (look like other types)
obj_tricky_values = {
    "word_true": "true",
    "word_false": "false",
    "word_null": "null",
    "word_True": "True",
    "word_None": "None",
}

# Mixed type list
obj_list_mixed = {
    "list_mixed": [1, "two", 3.0, True, False, None, {"a": 1}, [1, 2]],
}

# All test cases
memorydctn_objs = [
    obj_basic,
    obj_string_edge,
    obj_empty,
    obj_nested,
    obj_special_keys,
    obj_tricky_values,
    obj_list_mixed,
]



from typing import Protocol, Any, Callable, TypeVar, cast


class FlopFunc(Protocol):
    flops: int

    def __call__(self, *args: Any, **kwargs: Any) -> Any: ...


F = TypeVar("F", bound=Callable[..., Any])


def attach_flops(func: F, flops: int) -> FlopFunc:
    setattr(func, "flops", flops)
    return cast(FlopFunc, func)
